<?php
/**************************************************************************
sonos_event.php - Skript zum reagieren auf Abspielstatusaenderungen vom 
		  ZonePlayer mit dem Namen "Wohnzimmer" und direkten 
		  Antriggern von FHEM-Befehlen (In diesem Fall, um den 
		  entsprechenden Verstärker anzuschalten)

(c) by Reiner Leins (rleins@lmsoft.de)
***************************************************************************/


/* Das Skript wartet auf hereinkommende Verbindungsanforderungen, und sollte nicht automatisch beendet werden. */
set_time_limit(0);

//Zones-Datei einlesen und parsen
$zonelist_filename = dirname(__FILE__).'/zones.txt';
foreach(file($zonelist_filename) as $elem) {
        $elemparsed = explode('|', $elem);
        $zonelist[$elemparsed[0]]['IP'] = trim($elemparsed[1]);
        $zonelist[$elemparsed[0]]['UDN'] = trim($elemparsed[2]);
        $zonelist[$elemparsed[0]]['modelNumber'] = trim($elemparsed[3]);
}

unset($wohnzimmer_serviceProxy);
$formerPlayState = '';
$hasToStop = false;

/* Create the UPnP context */
$context = gupnp_context_new();
if (!$context) {
    printf("Error creating the GUPnP context\n");
    exit(-1);
}

/* Create the control point, searching for SwitchPower services */
$controlPoint = gupnp_control_point_new ($context, "urn:schemas-upnp-org:service:AVTransport:1");

/* Connect to the service-found callback */
$arg = array('zonelist' => $zonelist, 'controlPoint' => $controlPoint);
gupnp_control_point_callback_set($controlPoint, GUPNP_SIGNAL_SERVICE_PROXY_AVAILABLE, "service_proxy_available_cb", $arg);

/* Set timeout for automatic resubscription */
/* Sonos raeumt eine Subscription nach ca. 10 Minuten auf, damit keine Leichen in der nur 32 Listener
   langen Liste uebrig bleiben. Deshalb muessen wir den Timeout kurz genug einstellen, damit der automatismus von GUPnP
   eine Chance hat. */
gupnp_context_set_subscription_timeout($context, 480);

/* Start for browsing, can only be stopped by CTRL-C */
gupnp_control_point_browse_start($controlPoint);

// Subscribe beenden, leider kommt er nicht mehr hierher :-)
// Die Zeilen stehen hier nur als grundsaetzliche Idee, wie man das sauber beendet
gupnp_service_proxy_remove_notify($wohnzimmer_serviceProxy, 'LastChange');
gupnp_service_proxy_set_subscribed($wohnzimmer_serviceProxy, false);


/*****************************************************
 * Ab hier die Callback-Methoden
 *****************************************************/
function updateLastChange_cb($variable, $value, $arg) {
    global $hasToStop, $formerPlayState;

    $data = xml2array($value, 1, 'attribute');
    $trackdata = xml2array($data['Event']['InstanceID']['CurrentTrackMetaData']['attr']['val'], 1, 'attribute');

    //Temporaere Hilfsvariablen, damit die Zeilen nicht so lang werden
    unset($streamtitle);
    if (isset($trackdata['DIDL-Lite']['item']['r:streamContent']['value'])) {
    	$streamtitle = $trackdata['DIDL-Lite']['item']['r:streamContent']['value'];
    }

    unset($streamitemres);
    if (isset($trackdata['DIDL-Lite']['item']['res']['value'])) {
	$streamitemres = $trackdata['DIDL-Lite']['item']['res']['value'];
    }

    unset($transportstate);
    if (isset($data['Event']['InstanceID']['TransportState']['attr']['val'])) {
	$transportstate = $data['Event']['InstanceID']['TransportState']['attr']['val'];
    }


    //print_r($data);
    //print_r($trackdata);

    //Beim Verbindungsaufbau mit einem Streaming-Server oder im Stopped-Zustand keine Titel-Ausgabe machen; Nur einen Status...
    if ($transportstate == 'STOPPED'
	|| (isset($streamtitle) && $streamtitle == 'ZPSTR_CONNECTING')
	|| (isset($streamtitle) && $streamtitle == 'ZPSTR_BUFFERING')) {
	print "**********************************************************************************************************".
		"\nZeitstempel: ".date("Y-n-j H:i:s").
		"\nPlaystatus: ".$transportstate.
		"\nNo Title-Information to show!\n";
    } else {
    	//Titelausgabe
	if (isset($streamitemres) && substr($streamitemres, 0, 17) == 'x-sonosapi-stream') {
	    //Radiotitel
	    print "**********************************************************************************************************".
		"\nZeitstempel: ".date("Y-n-j H:i:s").
		"\nPlaystatus: ".$transportstate.
		"\nStream-Title: \"".(isset($streamtitle) ? $streamtitle : '-Not transferred yet-')."\"\n";
    	} else {
	    //Normaler Titel
	    print "**********************************************************************************************************".
		"\nZeitstempel: ".date("Y-n-j H:i:s").
		"\nPlaystatus: ".$transportstate.
		"\nTitle: (".$trackdata['DIDL-Lite']['item']['dc:creator']['value'].') "'.$trackdata['DIDL-Lite']['item']['dc:title']['value'].'" vom Album "'.$trackdata['DIDL-Lite']['item']['upnp:album']['value']."\"\n";
    	}
    }

    //Wenn der Abspiel-Status sich geaendert hat, FHEM entsprechend triggern
    if ($formerPlayState != $transportstate) {
	$formerPlayState = $transportstate;

	if ($formerPlayState == 'PLAYING') {
		system('perl /etc/fhem/fhem/fhem.pl 7072 "trigger SonosWebNanny_On"');
		print "Sonos-FHEM-Trigger: On\n";
	} else {
		system('perl /etc/fhem/fhem/fhem.pl 7072 "trigger SonosWebNanny_Off"');
		print "Sonos-FHEM-Trigger: Off\n";
	}
    }

    if ($hasToStop) {
	/* Stop browsing */
	gupnp_control_point_browse_stop($arg['controlPoint']);
    }
}

function service_proxy_available_cb($proxy, $arg)
{
    global $wohnzimmer_serviceProxy;

    $zonelist = $arg['zonelist'];

    $serviceInfo = gupnp_service_info_get($proxy);
    //print_r($serviceInfo);

    //Pruefen, ob das Wohnzimmer sich gemeldet hat :-)
    if ($serviceInfo['udn'] == $zonelist['Wohnzimmer']['UDN'].'_MR') {
	print "Wohnzimmer gefunden :-) ";

	$wohnzimmer_serviceProxy = $proxy;

	if (!gupnp_service_proxy_set_subscribed($proxy, true)) {
		print "Subscription failed!\n";
	}

	if (!gupnp_service_proxy_add_notify($proxy, 'LastChange', GUPNP_TYPE_STRING, 'updateLastChange_cb', $arg)) {
		print "Add Notify failed!\n";
	}
	print "und Callback eingerichtet!\n";
    }
}

/*****************************************************
 * Ab hier Hilfsfunktionen
 *****************************************************/

/**
 * xml2array() will convert the given XML text to an array in the XML structure.
 * Link: http://www.bin-co.com/php/scripts/xml2array/
 * Arguments : $contents - The XML text
 *                $get_attributes - 1 or 0. If this is 1 the function will get the attributes as well as the tag values - this results in a different array structure in the return value.
 *                $priority - Can be 'tag' or 'attribute'. This will change the way the resulting array sturcture. For 'tag', the tags are given more importance.
 * Return: The parsed XML in an array form. Use print_r() to see the resulting array structure.
 * Examples: $array =  xml2array(file_get_contents('feed.xml'));
 *              $array =  xml2array(file_get_contents('feed.xml', 1, 'attribute'));
 */
function xml2array($contents, $get_attributes=1, $priority = 'tag') {
    if(!$contents) return array();

    if(!function_exists('xml_parser_create')) {
        //print "'xml_parser_create()' function not found!";
        return array();
    }

    //Get the XML parser of PHP - PHP must have this module for the parser to work
    $parser = xml_parser_create('');
    xml_parser_set_option($parser, XML_OPTION_TARGET_ENCODING, "UTF-8"); # http://minutillo.com/steve/weblog/2004/6/17/php-xml-and-character-encodings-a-tale-of-sadness-rage-and-data-loss
    xml_parser_set_option($parser, XML_OPTION_CASE_FOLDING, 0);
    xml_parser_set_option($parser, XML_OPTION_SKIP_WHITE, 1);
    xml_parse_into_struct($parser, trim($contents), $xml_values);
    xml_parser_free($parser);

    if(!$xml_values) return;//Hmm...

    //Initializations
    $xml_array = array();
    $parents = array();
    $opened_tags = array();
    $arr = array();

    $current = &$xml_array; //Refference

    //Go through the tags.
    $repeated_tag_index = array();//Multiple tags with same name will be turned into an array
    foreach($xml_values as $data) {
        unset($attributes,$value);//Remove existing values, or there will be trouble

        //This command will extract these variables into the foreach scope
        // tag(string), type(string), level(int), attributes(array).
        extract($data);//We could use the array by itself, but this cooler.

        $result = array();
        $attributes_data = array();

        if(isset($value)) {
            if($priority == 'tag') $result = $value;
            else $result['value'] = $value; //Put the value in a assoc array if we are in the 'Attribute' mode
        }

        //Set the attributes too.
        if(isset($attributes) and $get_attributes) {
            foreach($attributes as $attr => $val) {
                if($priority == 'tag') $attributes_data[$attr] = $val;
                else $result['attr'][$attr] = $val; //Set all the attributes in a array called 'attr'
            }
        }

        //See tag status and do the needed.
        if($type == "open") {//The starting of the tag '<tag>'
            $parent[$level-1] = &$current;
            if(!is_array($current) or (!in_array($tag, array_keys($current)))) { //Insert New tag
                $current[$tag] = $result;
                if($attributes_data) $current[$tag. '_attr'] = $attributes_data;
                $repeated_tag_index[$tag.'_'.$level] = 1;

                $current = &$current[$tag];

            } else { //There was another element with the same tag name

                if(isset($current[$tag][0])) {//If there is a 0th element it is already an array
                    $current[$tag][$repeated_tag_index[$tag.'_'.$level]] = $result;
                    $repeated_tag_index[$tag.'_'.$level]++;
                } else {//This section will make the value an array if multiple tags with the same name appear together
                    $current[$tag] = array($current[$tag],$result);//This will combine the existing item and the new item together to make an array
                    $repeated_tag_index[$tag.'_'.$level] = 2;

                    if(isset($current[$tag.'_attr'])) { //The attribute of the last(0th) tag must be moved as well
                        $current[$tag]['0_attr'] = $current[$tag.'_attr'];
                        unset($current[$tag.'_attr']);
                    }

                }
                $last_item_index = $repeated_tag_index[$tag.'_'.$level]-1;
                $current = &$current[$tag][$last_item_index];
            }

        } elseif($type == "complete") { //Tags that ends in 1 line '<tag />'
            //See if the key is already taken.
            if(!isset($current[$tag])) { //New Key
                $current[$tag] = $result;
                $repeated_tag_index[$tag.'_'.$level] = 1;
                if($priority == 'tag' and $attributes_data) $current[$tag. '_attr'] = $attributes_data;

            } else { //If taken, put all things inside a list(array)
                if(isset($current[$tag][0]) and is_array($current[$tag])) {//If it is already an array...

                    // ...push the new element into that array.
                    $current[$tag][$repeated_tag_index[$tag.'_'.$level]] = $result;

                    if($priority == 'tag' and $get_attributes and $attributes_data) {
                        $current[$tag][$repeated_tag_index[$tag.'_'.$level] . '_attr'] = $attributes_data;
                    }
                    $repeated_tag_index[$tag.'_'.$level]++;

                } else { //If it is not an array...
                    $current[$tag] = array($current[$tag],$result); //...Make it an array using using the existing value and the new value
                    $repeated_tag_index[$tag.'_'.$level] = 1;
                    if($priority == 'tag' and $get_attributes) {
                        if(isset($current[$tag.'_attr'])) { //The attribute of the last(0th) tag must be moved as well

                            unset($current[$tag.'_attr']);
                        }

                        if($attributes_data) {
                            $current[$tag][$repeated_tag_index[$tag.'_'.$level] . '_attr'] = $attributes_data;
                        }
                    }
                    $repeated_tag_index[$tag.'_'.$level]++; //0 and 1 index is already taken
                }
            }

        } elseif($type == 'close') { //End of tag '</tag>'
            $current = &$parent[$level-1];
        }
    }

    return($xml_array);
}
?>
